"""
=============
psimulate CLI
=============

Command line interface for `psimulate`.

.. click:: vivarium_cluster_tools.psimulate.cli:psimulate
   :prog: psimulate
   :show-nested:

"""
from pathlib import Path
from typing import Optional

import click
from loguru import logger
from vivarium.framework.utilities import handle_exceptions

from vivarium_cluster_tools import cli_tools, logs
from vivarium_cluster_tools.psimulate import (
    COMMANDS,
    cluster,
    paths,
    redis_dbs,
    results,
    runner,
)
from vivarium_cluster_tools.psimulate.worker.load_test_work_horse import (
    get_psimulate_test_dict,
)


@click.group()
def psimulate():
    """A command line utility for running many simulations in parallel.

    You may initiate a new run with the ``run`` sub-command or restart a run
    from where it was stopped by using the ``restart`` sub-command.

    """
    pass


shared_options = [
    cluster.with_project,
    cluster.with_queue_and_max_runtime,
    cluster.with_peak_memory,
    redis_dbs.with_redis,
    results.with_no_batch,
    cli_tools.with_verbose_and_pdb,
]


@psimulate.command()
@click.argument(
    "model_specification",
    type=click.Path(exists=True, dir_okay=False),
    callback=cli_tools.coerce_to_full_path,
)
@click.argument(
    "branch_configuration",
    type=click.Path(exists=True, dir_okay=False),
    callback=cli_tools.coerce_to_full_path,
)
@click.option(
    "--artifact_path",
    "-i",
    type=click.Path(exists=True, dir_okay=False),
    help="The path to the artifact data file.",
    callback=cli_tools.coerce_to_full_path,
)
@click.option(
    "--result-directory",
    "-o",
    type=click.Path(file_okay=False),
    default=paths.DEFAULT_OUTPUT_DIRECTORY,
    show_default=True,
    help="The directory to write results to. A folder will be "
    "created in this directory with the same name as the "
    "configuration file.",
    callback=cli_tools.coerce_to_full_path,
)
@cli_tools.pass_shared_options(shared_options)
def run(
    model_specification: Path,
    branch_configuration: Path,
    artifact_path: Optional[Path],
    result_directory: Optional[Path],
    **options,
) -> None:
    """Run a parallel simulation.

    The simulation itself is defined by a MODEL_SPECIFICATION yaml file
    and the parameter changes across runs are defined by a BRANCH_CONFIGURATION
    yaml file.

    The path to the data artifact can be provided as an argument here, in the
    branch configuration, or in the model specification file. Values provided as
    a command line argument or in the branch specification file will override a
    value specified in the model specifications file. If an artifact path is
    provided both as a command line argument and to the branch configuration file
    a ConfigurationError will be thrown.

    Within the provided or default results directory, a subdirectory will be
    created with the same name as the MODEL_SPECIFICATION if one does not exist.
    Results will be written to a further subdirectory named after the start time
    of the simulation run.

    """
    logs.configure_main_process_logging_to_terminal(options["verbose"])
    main = handle_exceptions(runner.main, logger, options["with_debugger"])

    main(
        command=COMMANDS.run,
        input_paths=paths.InputPaths.from_entry_point_args(
            input_model_specification_path=model_specification,
            input_branch_configuration_path=branch_configuration,
            input_artifact_path=artifact_path,
            result_directory=result_directory,
        ),
        native_specification=cluster.NativeSpecification(
            job_name=model_specification.stem,
            project=options["project"],
            queue=options["queue"],
            peak_memory=options["peak_memory"],
            max_runtime=options["max_runtime"],
        ),
        redis_processes=options["redis"],
        no_batch=options["no_batch"],
        extra_args={},
    )


@psimulate.command()
@click.argument(
    "results-root",
    type=click.Path(exists=True, file_okay=False, writable=True),
    callback=cli_tools.coerce_to_full_path,
)
@cli_tools.pass_shared_options(shared_options)
def restart(results_root, **options):
    """Restart a parallel simulation from a previous run at RESULTS_ROOT.

    Restarting will not erase existing results, but will start workers to
    perform the remaining simulations.  RESULTS_ROOT is expected to be an
    output directory from a previous ``psimulate run`` invocation.

    """
    logs.configure_main_process_logging_to_terminal(options["verbose"])
    main = handle_exceptions(runner.main, logger, options["with_debugger"])

    main(
        command=COMMANDS.restart,
        input_paths=paths.InputPaths.from_entry_point_args(
            result_directory=results_root,
        ),
        native_specification=cluster.NativeSpecification(
            job_name=results_root.parent.name,
            project=options["project"],
            queue=options["queue"],
            peak_memory=options["peak_memory"],
            max_runtime=options["max_runtime"],
        ),
        redis_processes=options["redis"],
        no_batch=options["no_batch"],
        extra_args={},
    )


@psimulate.command()
@click.argument(
    "results-root",
    type=click.Path(exists=True, file_okay=False, writable=True),
    callback=cli_tools.coerce_to_full_path,
)
@click.option(
    "--add-draws",
    type=int,
    default=0,
    help="The number of input draws to add to a previous run.",
)
@click.option(
    "--add-seeds",
    type=int,
    default=0,
    help="The number of random seeds to add to a previous run.",
)
@cli_tools.pass_shared_options(shared_options)
def expand(results_root, **options):
    """Expand a previous run at RESULTS_ROOT by adding input draws and/or
    random seeds.

    Expanding will not erase existing results, but will start workers to perform
    the additional simulations determined by the added draws/seeds.
    RESULTS_ROOT is expected to be an output directory from a previous
    ``psimulate run`` invocation.

    """
    logs.configure_main_process_logging_to_terminal(options["verbose"])
    main = handle_exceptions(runner.main, logger, options["with_debugger"])

    main(
        command=COMMANDS.expand,
        input_paths=paths.InputPaths.from_entry_point_args(
            result_directory=results_root,
        ),
        native_specification=cluster.NativeSpecification(
            job_name=results_root.parent.name,
            project=options["project"],
            queue=options["queue"],
            peak_memory=options["peak_memory"],
            max_runtime=options["max_runtime"],
        ),
        redis_processes=options["redis"],
        no_batch=options["no_batch"],
        extra_args={
            "num_draws": options["add_draws"],
            "num_seeds": options["add_seeds"],
        },
    )


@psimulate.command()
@click.argument(
    "test-type",
    type=click.Choice(list(get_psimulate_test_dict())),
)
@click.option(
    "--num-workers",
    "-n",
    type=click.INT,
    default=1000,
)
@click.option(
    "--result-directory",
    "-o",
    type=click.Path(file_okay=False),
    default=f"{paths.DEFAULT_OUTPUT_DIRECTORY}/load_tests",
    callback=cli_tools.coerce_to_full_path,
)
@cli_tools.pass_shared_options(shared_options)
def test(test_type, num_workers, result_directory, **options):
    logs.configure_main_process_logging_to_terminal(options["verbose"])
    main = handle_exceptions(runner.main, logger, options["with_debugger"])
    main(
        command=COMMANDS.load_test,
        input_paths=paths.InputPaths.from_entry_point_args(
            result_directory=result_directory,
        ),
        native_specification=cluster.NativeSpecification(
            job_name=f"load_test_{test_type}",
            project=options["project"],
            queue=options["queue"],
            peak_memory=options["peak_memory"],
            max_runtime=options["max_runtime"],
        ),
        redis_processes=options["redis"],
        no_batch=options["no_batch"],
        extra_args={
            "test_type": test_type,
            "num_workers": num_workers,
        },
    )



"""
=====================================
Reproducible Programming Environments
=====================================

Tools to manage and validate reproducible Python environments for simulation runs.

"""
import os
import subprocess
from pathlib import Path
from typing import Dict, List, Tuple

from loguru import logger


def validate(environment_file: Path) -> None:
    pip_list_proc = subprocess.Popen(
        ["pip", "list", "--format=freeze"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    out, err = pip_list_proc.communicate()
    current_environment_list = out.decode().strip().split("\n")

    # Update permissions mask (assign to variable to avoid printing previous value)
    _ = os.umask(0o002)

    if not environment_file.exists():  # original run
        logger.info(f"Writing environment file to {str(environment_file)}.")
        environment_file.write_text("\n".join(current_environment_list))
    else:  # compare with original
        with environment_file.open() as f:
            original_environment_list = [line.replace("\n", "") for line in f]
        current_environment = _convert_pip_list_to_dict(current_environment_list)
        original_environment = _convert_pip_list_to_dict(original_environment_list)
        _compare_environments(current_environment, original_environment)
        logger.info(
            "Validation of environment successful. All pip installed packages match "
            "original versions. Run can proceed."
        )


def _parse_package_version(s: str) -> Tuple[str, str]:
    if (
        "no version control" in s
    ):  # installed from non-git controlled source code in editable mode
        s = s.split("(")[1].split(")")[0]  # yields <package>==<version>

    packaged = s.split("==")
    if len(packaged) == 2:  # not installed from source code, e.g., <package>==<version>
        package = packaged[0]
        version = packaged[1]
    else:  # installed from source code in editable mode so more parsing required
        sourced = s.split("=")
        package = sourced[-1]
        version = sourced[0].split("@")[-1].replace("#egg", "")  # version here is git hash
    return package, version


def _convert_pip_list_to_dict(pf_list: List[str]) -> Dict:
    return {p: v for p, v in [_parse_package_version(s) for s in pf_list]}


def _compare_environments(current: Dict, original: Dict) -> None:
    differences = []

    current_packages = set(current.keys())
    original_packages = set(original.keys())

    new_packages = current_packages.difference(original_packages)
    if new_packages:
        differences.append(
            f"The current environment contains the following packages not present "
            f"in the original environment: {new_packages}."
        )

    missing_packages = original_packages.difference(current_packages)
    if missing_packages:
        differences.append(
            f"The current environment is missing the following packages present in the original "
            f"environment: {missing_packages}."
        )

    differing_versions = []
    for p in current_packages.intersection(original_packages):
        if current[p] != original[p]:
            differing_versions.append(f"{p}: {original[p]} -> {current[p]}")

    if differing_versions:
        differences.append(
            f"Different versions found in current and original environment for the following "
            f"packages: {differing_versions}."
        )

    if differences:
        differences = "\n".join(differences)
        raise ValueError(
            f"Differences found between environment used for original run and current environment. "
            f"In order to successfully run, you should make a new environment using the requirements.txt "
            f"file found in the output directory. Differences found as follows: {differences}."
        )



"""
==============================
Branch and Keyspace Management
==============================

Tools for managing the parameter space of a parallel run.

"""
from itertools import product
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Tuple, Union

import numpy as np
import yaml
from vivarium.framework.utilities import collapse_nested_dict

from vivarium_cluster_tools.psimulate.model_specification import FULL_ARTIFACT_PATH_KEY


class Keyspace:
    """A representation of a collection of simulation configurations."""

    def __init__(self, branches: List[Dict], keyspace: Dict):
        self.branches = branches
        self._keyspace = keyspace

    @classmethod
    def from_branch_configuration(
        cls, branch_configuration_file: Union[str, Path]
    ) -> "Keyspace":
        """
        Parameters
        ----------
        branch_configuration_file
            Absolute path to the branch configuration file.
        """
        branches, input_draw_count, random_seed_count = load_branch_configuration(
            branch_configuration_file
        )
        keyspace = calculate_keyspace(branches)
        keyspace["input_draw"] = calculate_input_draws(input_draw_count)
        keyspace["random_seed"] = calculate_random_seeds(random_seed_count)

        return Keyspace(branches, keyspace)

    @classmethod
    def from_previous_run(cls, keyspace_path: Path, branches_path: Path) -> "Keyspace":
        keyspace = yaml.full_load(keyspace_path.read_text())
        branches = yaml.full_load(branches_path.read_text())
        return Keyspace(branches, keyspace)

    @classmethod
    def from_entry_point_args(
        cls,
        input_branch_configuration_path: Optional[Path],
        keyspace_path: Path,
        branches_path: Path,
        extras: Dict,
    ) -> "Keyspace":
        if input_branch_configuration_path is not None:
            keyspace = cls.from_branch_configuration(
                input_branch_configuration_path,
            )
        elif keyspace_path.exists():
            keyspace = cls.from_previous_run(keyspace_path, branches_path)
            keyspace.add_draws(extras.get("num_draws", 0))
            keyspace.add_seeds(extras.get("num_seeds", 0))
        else:
            keyspace = Keyspace([], {})

        return keyspace

    def persist(self, keyspace_path: Path, branches_path: Path) -> None:
        keyspace_path.write_text(yaml.dump(self._keyspace))
        branches_path.write_text(yaml.dump(self.branches))

    def add_draws(self, num_draws: int) -> None:
        existing = self._keyspace["input_draw"]
        additional = calculate_input_draws(num_draws, existing)
        self._keyspace["input_draw"] = existing + additional

    def add_seeds(self, num_seeds: int) -> None:
        existing = self._keyspace["random_seed"]
        additional = calculate_random_seeds(num_seeds, existing)
        self._keyspace["random_seed"] = existing + additional

    def __contains__(self, item: str) -> bool:
        """Checks whether the item is present in the Keyspace"""
        return item in self._keyspace

    def __iter__(self) -> Iterator[Tuple[int, int, Dict]]:
        """Yields and individual simulation configuration from the keyspace."""
        for job_config in product(
            self._keyspace["input_draw"], self._keyspace["random_seed"], self.branches
        ):
            if job_config[2] is None:
                job_config[2] = {}
            yield job_config

    def __len__(self) -> int:
        """Returns the number of individual simulation runs this keyspace represents."""
        return len(
            list(
                product(
                    self._keyspace["input_draw"], self._keyspace["random_seed"], self.branches
                )
            )
        )


def calculate_input_draws(
    input_draw_count: int, existing_draws: List[int] = None
) -> List[int]:
    """Determines a random sample of the GBD input draws to use given a draw
    count and any existing draws.

    Parameters
    ----------
    input_draw_count
        The number of draws to pull.

    existing_draws
        Any draws that have already been pulled and should not be pulled again.

    Returns
    -------
    List[int]
        A set of unique input draw numbers, guaranteed not to overlap with any
        existing draw numbers.

    """
    max_draw_count = 1000
    if input_draw_count > max_draw_count:
        raise ValueError(f"Input draw count must be less than {max_draw_count}.")

    possible = list(range(max_draw_count))
    if existing_draws:
        possible = sorted(list(set(possible).difference(existing_draws)))
        min_input_draw_count_allowed = 0
    else:
        min_input_draw_count_allowed = 1
    np.random.seed(123456)
    np.random.shuffle(possible)
    if min_input_draw_count_allowed <= input_draw_count <= len(possible):
        return possible[:input_draw_count]
    else:
        raise ValueError(
            f"Input draw count must be between {min_input_draw_count_allowed} "
            f"and {len(possible)} (inclusive). You specified {input_draw_count}."
        )


def calculate_random_seeds(
    random_seed_count: int, existing_seeds: List[int] = None
) -> List[int]:
    """Generates random seeds to use given a count of seeds and any existing
    seeds.

    Parameters
    ----------
    random_seed_count
        The number of random seeds to generate.

    existing_seeds
        Any random seeds that have already been generated and should not be
        generated again.

    Returns
    -------
    List[int]
        A set of unique random seeds, guaranteed not to overlap with any
        existing random seeds.

    """
    if not random_seed_count:
        return []

    max_seed_count = 10000
    if random_seed_count > max_seed_count:
        raise ValueError(f"Random seed count must be less than {max_seed_count}.")

    possible = list(range(max_seed_count))
    if existing_seeds:
        possible = sorted(list(set(possible).difference(existing_seeds)))
    np.random.seed(654321)
    np.random.shuffle(possible)
    return possible[:random_seed_count]


def calculate_keyspace(branches: List[Dict]) -> Dict:
    keyspace = {k: {v} for k, v in collapse_nested_dict(branches[0])}

    for branch in branches[1:]:
        branch = dict(collapse_nested_dict(branch))
        if set(branch.keys()) != set(keyspace.keys()):
            raise ValueError("All branches must have the same keys")
        for k, v in branch.items():
            if k == FULL_ARTIFACT_PATH_KEY:
                validate_artifact_path(v)
            keyspace[k].add(v)
    keyspace = {k: list(v) for k, v in keyspace.items()}
    return keyspace


def load_branch_configuration(path: Path) -> Tuple[List[Dict], int, int]:
    data = yaml.full_load(path.read_text())

    input_draw_count = data.get("input_draw_count", 1)
    random_seed_count = data.get("random_seed_count", 1)

    assert input_draw_count <= 1000, "Cannot use more that 1000 draws from GBD"

    if "branches" in data:
        branches = expand_branch_templates(data["branches"])
    else:
        branches = [{}]

    return branches, input_draw_count, random_seed_count


def expand_branch_templates(templates: Dict) -> List[Dict]:
    """
    Take a list of dictionaries of configuration values (like the ones used in
    experiment branch configurations) and expand it by taking any values which
    are lists and creating a new set of branches which is made up of the
    product of all those lists plus all non-list values.

    For example this:

    .. code::

       {'a': {'b': [1,2], 'c': 3, 'd': [4,5,6]}}

    becomes this:

    .. code::

       [
           {'a': {'b': 1, 'c': 3, 'd': 4}},
           {'a': {'b': 2, 'c': 3, 'd': 5}},
           {'a': {'b': 1, 'c': 3, 'd': 6}},
           {'a': {'b': 2, 'c': 3, 'd': 4}},
           {'a': {'b': 1, 'c': 3, 'd': 5}},
           {'a': {'b': 2, 'c': 3, 'd': 6}}
       ]

    """
    expanded_branches = []

    for branch in templates:
        branch = sorted(collapse_nested_dict(branch))
        branch = [(k, v if isinstance(v, list) else [v]) for k, v in branch]
        expanded_size = np.product([len(v) for k, v in branch])
        new_branches = []
        pointers = {k: 0 for k, _ in branch}
        for _ in range(expanded_size):
            new_branch = []
            tick = True
            for k, v in branch:
                new_branch.append((k, v[pointers[k]]))
                if tick:
                    i = pointers[k] + 1
                    if i < len(v):
                        tick = False
                        pointers[k] = i
                    else:
                        pointers[k] = 0
            new_branches.append(new_branch)
        expanded_branches.extend(new_branches)

    final_branches = []
    for branch in expanded_branches:
        root = {}
        final_branches.append(root)
        for k, v in branch:
            current = root
            *ks, k = k.split(".")
            for sub_k in ks:
                if sub_k in current:
                    current = current[sub_k]
                else:
                    current[sub_k] = {}
                    current = current[sub_k]
            current[k] = v

    return final_branches


def validate_artifact_path(artifact_path: str) -> str:
    """Validates that the path to the data artifact from the branches file exists.

    The path specified in the configuration must be absolute

    Parameters
    ----------
    artifact_path
        The path to the artifact.

    """
    path = Path(artifact_path)

    if not path.is_absolute() or not path.exists():
        raise FileNotFoundError(f"Cannot find artifact at path {path}")

    return str(path)



"""
==================
Results Processing
==================

Tools for processing and writing results.

"""
import time
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
from loguru import logger

from vivarium_cluster_tools import utilities as vct_utils


def write_results_batch(
    output_directory: Path,
    written_results: pd.DataFrame,
    unwritten_results: List[pd.DataFrame],
    batch_size: int = 50,
) -> Tuple[pd.DataFrame, List[pd.DataFrame]]:
    new_results_to_write, unwritten_results = (
        unwritten_results[:batch_size],
        unwritten_results[batch_size:],
    )
    results_to_write = _concat_results(written_results, new_results_to_write)

    start = time.time()
    _safe_write_results(results_to_write, output_directory / "output.hdf")
    end = time.time()
    logger.info(f"Updated output.hdf in {end - start:.4f}s.")
    return results_to_write, unwritten_results


def _concat_results(
    old_results: pd.DataFrame, new_results: List[pd.DataFrame]
) -> pd.DataFrame:
    # Skips all the pandas index checking because columns are in the same order.
    start = time.time()

    to_concat = [d.reset_index(drop=True) for d in new_results]
    if not old_results.empty:
        to_concat += [old_results.reset_index(drop=True)]

    results = _concat_preserve_types(to_concat)

    end = time.time()
    logger.info(f"Concatenated {len(new_results)} results in {end - start:.2f}s.")
    return results


def _concat_preserve_types(df_list: List[pd.DataFrame]) -> pd.DataFrame:
    """Concatenation preserves all ``numpy`` dtypes but does not preserve any
    pandas specific dtypes (e.g., categories become objects."""
    dtypes = df_list[0].dtypes
    columns_by_dtype = [list(dtype_group.index) for _, dtype_group in dtypes.groupby(dtypes)]

    splits = []
    for columns in columns_by_dtype:
        slices = [df.filter(columns) for df in df_list]
        splits.append(pd.DataFrame(data=np.concatenate(slices), columns=columns))
    return pd.concat(splits, axis=1)


@vct_utils.backoff_and_retry(backoff_seconds=30, num_retries=3, log_function=logger.warning)
def _safe_write_results(results: pd.DataFrame, output_path: Union[str, Path]) -> None:
    # Writing to a hdf over and over balloons the file size so
    # write to new file and move it over to avoid
    temp_output_path = output_path.with_name(output_path.name + "update")
    results.to_hdf(temp_output_path, "data")
    temp_output_path.replace(output_path)



"""
==================
Results Management
==================

"""
from vivarium_cluster_tools.psimulate.results.cli_options import (
    with_no_batch,
    with_no_cleanup,
)
from vivarium_cluster_tools.psimulate.results.processing import write_results_batch



"""
===================
Cluster Interaction
===================

Tools for interacting with the IHME cluster.

"""
from vivarium_cluster_tools.psimulate.cluster.cli_options import (
    with_peak_memory,
    with_project,
    with_queue_and_max_runtime,
)
from vivarium_cluster_tools.psimulate.cluster.interface import (
    NativeSpecification,
    submit_worker_jobs,
    validate_cluster_environment,
)
